# fundamental functions for the det_gamma calibration by the wire grid
# Please contact hnakata_JP(zhongtianjiaxin@gmail.com) if you have questions.

# import python standard libraries
import numpy as np
from dataclasses import dataclass
from typing import Union
from scipy.optimize import least_squares
from scipy.interpolate import interp1d

# SO specific
from so3g.hk import load_range
from sotodlib import core
from sotodlib.io import hkdb
from sotodlib.io import hk_utils
from sotodlib.site_pipeline import util
logger = util.init_logger('wiregrid', 'wiregrid: ')

# hardware specific constants, which may have to be sperated as a config file
WG_COUNTS2RAD = 2*np.pi/52000 # SATP1~SATP3, JSAT(SATP4?) will be 104000
WG_OFFSET_SATP1 = np.deg2rad(180-(47.37+33.5+87)) # SAT1, MF1, which is 12.13 deg
WG_OFFSET_SATP2 = np.deg2rad(180-(50.127+33.5+87)) # SAT2, UHF, which is 9.473 deg
WG_OFFSET_SATP3 = np.deg2rad(180-(48.29+33.5+87)) # TSAT, MF2, which is 11.21 deg

@dataclass
class L2Config:
    path: str
    telescope: str
    start_time: float
    stop_time: float
    timestamp_margin: float = 1.0

@dataclass
class L3Config:
    path: str
    start_time: float
    stop_time: float

DataSourceConfig = Union[L2Config, L3Config]

def load_data(config: DataSourceConfig) -> dict:
    """
    Load wire grid house-keeping data based on the provided configuration.
    start_time and stop_time are used to load the data within the specified range.

    The L3Config file path should point the sqlite database file.
    The L2Config file path should point the directory where the house-keeping data is stored.

    Notes
    -----
        We assume wg_cfg is a yaml file, which has the following structure.

        - hk_root. path to house-keeping data directory, e.g. ``'../data/satp1/hk/'``
        - db_file.  path to sqlite database file, e.g. ``'./hkdb-satp1.db'``
        - aliases:
            - enc_rad_raw : ``'wg-encoder.wgencoder_full.reference_count'``
            - LSL1 : ``'wg-actuator.wgactuator.limitswitch_LSL1'``
            - LSL2 : ``'wg-actuator.wgactuator.limitswitch_LSL2'``
            - LSR1 : ``'wg-actuator.wgactuator.limitswitch_LSR1'``
            - LSR2 : ``'wg-actuator.wgactuator.limitswitch_LSR2'``
            - angleX : ``'wg-tilt-sensor.wgtiltsensor.angleX'``
            - angleY : ``'wg-tilt-sensor.wgtiltsensor.angleY'``
            - tempX : ``'wg-tilt-sensor.wgtiltsensor.tempX'``
            - tempY : ``'wg-tilt-sensor.wgtiltsensor.tempY'``
    """
    if isinstance(config, L2Config):
        logger.info("Loading the wire grid house-keeping data with hk_dir.")
        return load_l2_data(config)
    elif isinstance(config, L3Config):
        logger.info("Loading the wire grid house-keeping data with hkdb.")
        return load_l3_data(config)
    else:
        raise ValueError("Unsupported config type")

def load_l2_data(config: L2Config) -> dict:
    """
    Load wire grid house-keeping data with so3g.hk.load_range method.
    """
    aliases = [
        'enc_rad_raw',
        'LSL1', 'LSL2', 'LSR1', 'LSR2',
        'angleX', 'angleY', 'tempX', 'tempY'
    ]
    fields = [
        f'{config.telescope}.wg-encoder.feeds.wgencoder_full.reference_count',
        f'{config.telescope}.wg-actuator.feeds.wgactuator.limitswitch_LSL1',
        f'{config.telescope}.wg-actuator.feeds.wgactuator.limitswitch_LSL2',
        f'{config.telescope}.wg-actuator.feeds.wgactuator.limitswitch_LSR1',
        f'{config.telescope}.wg-actuator.feeds.wgactuator.limitswitch_LSR2',
        f'{config.telescope}.wg-tilt-sensor.feeds.wgtiltsensor.angleX',
        f'{config.telescope}.wg-tilt-sensor.feeds.wgtiltsensor.angleY',
        f'{config.telescope}.wg-tilt-sensor.feeds.wgtiltsensor.temperatureX',
        f'{config.telescope}.wg-tilt-sensor.feeds.wgtiltsensor.temperatureY',
    ]
    _start = config.start_time - config.timestamp_margin  # margin of 1 second
    _stop = config.stop_time + config.timestamp_margin
    _data = load_range(
        _start, _stop, data_dir=config.path,
        fields=fields,
        alias=aliases,
    )

    if len(_data) == 0:
        logger.error("No available wire grid data with this tod.")
        raise ValueError("No available wire grid data with this tod.")

    if 'enc_rad_raw' not in _data:
        logger.error("No available ENCODER data with this tod.")
        raise ValueError("No available ENCODER data with this tod.")

    if 'angleX' not in _data or \
        'angleY' not in _data or \
        'tempX' not in _data or \
        'tempY' not in _data:
        logger.warning("Lack of tilt sensor data.")

    if 'LSL1' not in _data or \
        'LSL2' not in _data or \
        'LSR1' not in _data or \
        'LSR2' not in _data:
        logger.warning("The actuator data is not correctly stored in house-keeping. \
                        \n inside/outside status is not certificated so far.")
    raw_data_dict = _data
    _in_radians = raw_data_dict['enc_rad_raw'][1] * WG_COUNTS2RAD
    raw_data_dict['enc_rad_raw'] = (
    raw_data_dict['enc_rad_raw'][0], _in_radians
        )
    return raw_data_dict

def load_l3_data(config: L3Config) -> dict:
    cfg = hkdb.HkConfig.from_yaml(config.path)
    lspec = hkdb.LoadSpec(
        cfg=cfg, start=config.start_time, end=config.stop_time,
        fields=list(cfg.aliases.keys())
    )
    raw_data = hkdb.load_hk(lspec)
    _fields = list(raw_data.data.keys())
    _swapped_aliases = {v: k for k, v in cfg.aliases.items()}
    _aliases = [_swapped_aliases[f] for f in _fields]
    raw_data_dict = {
        alias: raw_data.data[field] for alias, field in zip(_aliases, _fields)
    }
    _in_radians = raw_data_dict['enc_rad_raw'][1] * WG_COUNTS2RAD
    raw_data_dict['enc_rad_raw'] = (
    raw_data_dict['enc_rad_raw'][0], _in_radians
        )
    return raw_data_dict

# Wrap house-keeping data
def wrap_wg_hk(tod, raw_data_dict, merge=True):
    """
    Wrap the house-keeping data about the wire grid operation.

    Parameters
    ----------
        tod : AxisManager

    Returns
    -------
        tod : AxisManager
            This includes fields, which are related with the wire grid hardware.

                - enc_rad_raw : wires' direction read by encoder in radian (raw data from the encoder)
                - LSL1 : ON/OFF status of the limit switch LEFT 1 (outside) of the actuator
                - LSL2 : ON/OFF status of the limit switch LEFT 2 (inside) of the actuator
                - LSR1 : ON/OFF status of the limit switch RIGHT 1 (outside) of the actuator
                - LSR2 : ON/OFF status of the limit switch RIGHT 2 (inside) of the actuator
                - angleX : tilt angle X of the tilt sensor in degree
                - angleY : tilt angle Y of the tilt sensor in degree
                - tempX : temperature sensor X of the tilt sensor in Celsius
                - tempY : temperature sensor Y of the tilt sensor in Celsius

    """
    if not hasattr(tod, 'timestamps'):
        logger.warning("TOD does not have timestamps, and it should be only meta data. \
                        \n Calculating TOD samplings under 5 ms samples. \
                        \n The sampling of this wrapped data is not correct.")
        _start = tod.obs_info.start_time
        _stop = tod.obs_info.stop_time
        timestamps = np.linspace(_start, _stop, tod.samps.count)
    else:
        timestamps = tod.timestamps

    interp_func = {}
    for _k, _v in raw_data_dict.items():
        interp_func[_k] = interp1d(_v[0], _v[1], bounds_error=False)

    hk_aman = core.AxisManager(core.OffsetAxis('samps', count=len(timestamps)))
    for _f in interp_func.keys():
        _cosamp_data = interp_func[_f](timestamps)
        hk_aman.wrap(_f, _cosamp_data, [(0, 'samps')])
    if merge:
        _ax_instrument = core.AxisManager(tod.samps)
        _ax_instrument.wrap('instrument', hk_aman)
        tod.wrap('wg', _ax_instrument)
        return tod
    return hk_aman

# Correct wires' direction for each telescope
def correct_wg_angle(tod):
    """
    Correct offset of wires' direction with value confirmed in the hardware testing.

    Parameters
    ----------
        tod : AxisManager

    """
    _tel = tod.obs_info.telescope
    if _tel == 'satp1':
        wg_offset = WG_OFFSET_SATP1
    elif _tel == 'satp2':
        wg_offset = WG_OFFSET_SATP2
    elif _tel == 'satp3':
        wg_offset = WG_OFFSET_SATP3
    else:
        logger.warning(f"No matched telescope name of {_tel} for wire grid offset value, wg_offset.\n \
                        the encoder offset is specified to zero. The hardware offset must remain.")
        wg_offset = 0
    enc_rad = (-tod.wg.instrument.enc_rad_raw + wg_offset)%(2*np.pi)
    tod.wg.instrument.wrap('enc_rad', enc_rad, [(0, 'samps')])
    return True

# Detect the motion of the wire grid
def _detect_motion(count, flag=0):
    """
    Detect the rotation of the grid.

    Parameters
    ----------
        count : int
            encoder actual counts
        flag : int
            flag to choose moving or static

    Returns
    -------
        count: int
            counts in the same status related with the flag

    """
    if flag == 0:
        vecbool = count[1:]-count[:-1] != 0  # moving
    else:
        vecbool = count[1:]-count[:-1] == 0  # static
    lvec = len(vecbool)
    inv_vec = vecbool[::-1]
    left = np.maximum.accumulate(
        list(map(lambda x: 0 if vecbool[x] == True else x, np.arange(lvec)))
    )
    right = lvec - 1 - np.maximum.accumulate(
        list(map(lambda x: 0 if inv_vec[x] == True else x, np.arange(lvec)))
    )[::-1]
    return right-left

# Define the timestamp and the wire angle in each step
def _detect_steps(tod, steps_thresholds=(10, 300)):
    """
    Detect all steps of the stepwise operation

    Parameters
    ----------
        tod : AxisManager
        steps_thresholds : Tuple
            the thresholds on the encoder counts.
            the first element is the upper bound for the static state,
            and the second element is the lower bound for the difference of the first

    Returns
    -------
        idx_steps_start, idx_steps_stop : sample index of the start/stop for each step

    """
    cdiff0 = np.where(_detect_motion(tod.wg.instrument.enc_rad/WG_COUNTS2RAD) < steps_thresholds[0], 1, 0)
    cdiff1 = np.where(_detect_motion(cdiff0, flag=1) > steps_thresholds[1], 0, 1)
    cdiff2 = cdiff1[1:] - cdiff1[:-1]
    idx_step_stop = np.where(cdiff2 == 1)[0]
    idx_step_start = np.where(cdiff2 == -1)[0]

    if len(idx_step_start) == len(idx_step_stop):
        logger.info(
            "The count of start index is equal to that of stop index."
        )
        step_size = np.average(idx_step_stop[1:] - idx_step_start[:-1]).astype(int)
        idx_step_stop = np.append(idx_step_stop, idx_step_start[-1]+step_size)

    elif len(idx_step_start) == len(idx_step_stop) + 1:
        logger.warning(
            "The count of start index is one greater than that of stop index. \
            \n Trying to fix the index by the average step size."
            )
        if idx_step_start[0] < idx_step_stop[0]:
            idx_step_start = idx_step_start[1:]
        # the other case has to be implemented in the future
        step_size = np.average(idx_step_stop[1:] - idx_step_start[:-1]).astype(int)
        idx_step_stop = np.append(idx_step_stop, idx_step_start[-1]+step_size)

    elif len(idx_step_start) + 1 == len(idx_step_stop):
        if idx_step_start[0] < idx_step_stop[0]:
            logger.warning(
                "The count of stop index is one less than that of start index. \
                \n Pass the original index to the next step."
            )
        else:
            logger.warning("_detect_steps: Future subject. Not corrected yet.")

    return idx_step_start, idx_step_stop

# Find the nominal operation range of the wire grid
def find_operation_range(tod, steps_thresholds=(10, 300), ls_margin=2000, is_restrict=True, \
                        remove_trembling=True, tremble_threshold=1.):
    """
    Define the range of the wire grid operation by the limit switches

    Parameters
    ----------
        tod : AxisManager
        ls_margin : int (default, 2000 samples)
            the offsets to define the operation range of the wire grid calibration.
        is_restrict : bool (default, True)
            whether restrict TODs by the opration range or not
        remove_trembling : bool (default, True)
            whether remove the steps in the encoder data under the motor-motion threshold, tremble_threshold
        tremble_threshold : float (default, 1 deg)
            the threshold to remove the slight vibration in the steps of the wire grid rotation

    Returns
    -------
        idx_steps_start, idx_steps_stop : sample index of the start/stop for each step

    """

    if ('LSR2' in tod.wg.instrument) or ('LSL2' in tod.wg.instrument):
        _r2_nan = np.isnan(tod.wg.instrument.LSR2)
        _l2_nan = np.isnan(tod.wg.instrument.LSL2)
        _ls_nan = _r2_nan & _l2_nan
        ls_state = np.where(
            _ls_nan, 0,
            tod.wg.instrument.LSR2.astype(int) | tod.wg.instrument.LSL2.astype(int)
        )
        _idx_temp = np.where(ls_state == 1)[0]
        idx_wg_oper = np.arange(
            _idx_temp[0] - ls_margin if 0 < _idx_temp[0] - ls_margin else 0,
            _idx_temp[-1] + ls_margin
            )

    else:
        logger.warning("Define the oparation range by the encoder data instead of limit switches. \
                        \n This module doesn't guarantee that the wire grid was inside.")
        # detect typical motion of the calibrator just in case
        try:
            _idx_step_start, _ = _detect_steps(tod, steps_thresholds=steps_thresholds)
        except IndexError:
            logger.error("The rotation couldn't be detected in this obs.")
        _ts_margin = 30
        _before_motion = np.where(tod.timestamps < tod.timestamps[_idx_step_start][0] - _ts_margin)[0]
        _after_motion  = np.where(tod.timestamps[_idx_step_start][-1] + _ts_margin < tod.timestamps)[0]
        idx_cal_start = max(_before_motion) if len(_before_motion) != 0 else 0
        idx_cal_stop = min(_after_motion) if len(_after_motion) != 0 else -1
        idx_wg_oper = np.arange(idx_cal_start, idx_cal_stop)

    if is_restrict:
        tod.restrict('samps', (idx_wg_oper[0], idx_wg_oper[-1]), in_place=True)

    # detect typical motion of the calibrator
    idx_steps_start, idx_steps_stop = _detect_steps(tod, steps_thresholds=steps_thresholds)

    # remove the rattling in the detected steps with the threshold of tremble_threshold
    if remove_trembling:
        _theta = tod.wg.instrument.enc_rad
        _idx_start = idx_steps_start
        _idx_stop = idx_steps_stop[1:]

        _fluc_start = np.where(np.abs(np.diff(_theta[_idx_start])) < np.deg2rad(tremble_threshold))[0]
        _fluc_stop = np.where(np.abs(np.diff(_theta[_idx_start])) < np.deg2rad(tremble_threshold))[0]
        assert _fluc_start.all() == _fluc_stop.all()

        idx_start_real = np.delete(_idx_start, _fluc_start+1)
        idx_stop_real = np.delete(_idx_stop, _fluc_stop)
        return idx_start_real, idx_stop_real

    return idx_steps_start, idx_steps_stop[1:]

# Wrap the data specified from the operation range of the wire grid
def calc_calibration_data_set(tod, idx_steps_start, idx_steps_stop):
    """
    Interpret the Q_cal, U_cal data from the wrapped calibration data.
    This method is based on both of demodQ and demodU, which are calculated by the HWP demodulation process.
    Users have to apply some HWP process before calling this.

    Parameters
    ----------
        tod : AxisManager,
            the others are same as the _get_operation_range
        idx_steps_start : np.ndarray
            sample index of the start of each step
        idx_steps_stop : np.ndarray
            sample index of the stop of each step

    Returns
    -------
        axes : AxisManger
        return includes fields as follows:

            - wg.cal_data.theta_wire_rad is the average direction of wires in each step
            - wg.cal_data.theta_wire_std is the standard deviation of the direction of wires in each step
            - wg.cal_data.ts_step_mid is the time stamps in the middle of each step
            - wg.cal_data.Q, wg.cal_data.U : Q (and U) signal by wires
            - wg.cal_data.Qerr, wg.cal_data.Uerr : the standard deviations of Q (and U) signal
    """
    _theta_wire = tod.wg.instrument.enc_rad

    ts_step_mid = []
    theta_wire_av = []
    theta_wire_std = []
    step_Q = []
    step_U = []
    step_Qerr = []
    step_Uerr = []
    for _i, _ in enumerate(idx_steps_start):
        _idx_frame = np.arange(tod.samps.count)
        instep = np.where(idx_steps_start[_i] < _idx_frame, True, False)
        instep = np.where(_idx_frame < idx_steps_stop[_i], instep, False)
        ts_step_mid.append(np.average(tod.timestamps[instep]))
        theta_wire_av.append(np.average(_theta_wire[instep]))
        theta_wire_std.append(np.std(_theta_wire[instep]))
        step_Q.append(np.average(tod.demodQ, axis=1, weights=instep))
        step_U.append(np.average(tod.demodU, axis=1, weights=instep))
        step_Qerr.append(np.std(tod.demodQ[:, instep == True], axis=1))
        step_Uerr.append(np.std(tod.demodU[:, instep == True], axis=1))
    ts_step_mid = np.array(ts_step_mid)
    theta_wire_av = np.array(theta_wire_av)
    theta_wire_std = np.array(theta_wire_std)
    step_Q = np.array(step_Q)
    step_U = np.array(step_U)
    step_Qerr = np.array(step_Qerr)
    step_Uerr = np.array(step_Uerr)

    # wrap the data for each step
    if 'cal_data' in tod.wg._fields.keys():
        tod.wg.move('cal_data', None)
    ax = core.AxisManager(tod.samps, tod.dets, core.IndexAxis('wg_steps'))
    ax.wrap('idx_steps_start', idx_steps_start, [(0, 'wg_steps')])
    ax.wrap('idx_steps_stop',  idx_steps_stop,  [(0, 'wg_steps')])
    ax.wrap('theta_wire_rad', theta_wire_av , [(0, 'wg_steps')])
    ax.wrap('theta_wire_std', theta_wire_std, [(0, 'wg_steps')])
    ax.wrap('ts_step_mid',    ts_step_mid,    [(0, 'wg_steps')])
    ax.wrap('Q',              step_Q,         [(0, 'wg_steps'), (1, 'dets')])
    ax.wrap('U',              step_U,         [(0, 'wg_steps'), (1, 'dets')])
    ax.wrap('Qerr',           step_Qerr,      [(0, 'wg_steps'), (1, 'dets')])
    ax.wrap('Uerr',           step_Uerr,      [(0, 'wg_steps'), (1, 'dets')])
    tod.wg.wrap('cal_data', ax)
    return ax

### Circle fitting functions start from here ###
def _circle_model(params, x, y, xerr, yerr):
    cx, cy, r = params
    dx = x - cx
    dy = y - cy
    rho = np.sqrt(dx**2 + dy**2) + 1e-15 # to avoid zero division
    model = rho - r
    sigma_m2 = (dx/rho)**2 * xerr**2 + (dy/rho)**2 * yerr**2
    return model / np.sqrt(sigma_m2)

def fit_with_circle(tod):
    """
    Get the results by the circle fitting about the responce against the wires in Q+iU plane

    Parameters
    ----------
        tod : AxisManager

    Returns
    -------
        fit_results : list
            the result of the circle fitting of the wires' signal in the Q/U plane.

    Notes
    -----
        With this process, tod will include the several parameters of the fittings as tod.wg.cfit_result:

            - cx0(_err) : the estimated x-offset value(, and its fit err).
            - cy0(_err) : the estimated y-offset value(, and its fit err).
            - cr(_err) : Estimated radius vaule(, and its fit err).
            - covariance : covariance matrix of the estimated parameters.
            - residual_var : Residual variance.
            - is_success : the status of how fits end.

    """
    _cal = tod.wg.cal_data
    fit_results = []
    params_val = []
    params_err = []
    covs = []
    res_vars = []
    is_success = []

    for _i in range(tod.dets.count):
        _obs_data = np.array([_cal.Q[:, _i], _cal.U[:, _i]])
        _obs_std = np.array([_cal.Qerr[:, _i], _cal.Uerr[:, _i]])

        # initial parameters
        cx0 = np.nanmean(_obs_data[0])
        cy0 = np.nanmean(_obs_data[1])
        cr = np.sqrt(np.nanmean(_obs_data[0]**2 + _obs_data[1]**2))
        _init_params = np.array([cx0, cy0, cr])
        _bounds = ([-2*cr, -2*cr, 0], [2*cr, 2*cr, 10*cr])

        # fit the data
        out = least_squares(
            _circle_model, _init_params, bounds=_bounds,
            args=(_obs_data[0], _obs_data[1], _obs_std[0], _obs_std[1])
        )
        fit_results.append(out)

        # Jacobian matrix and covariance matrix
        J = out.jac
        RSS = float(out.fun @ out.fun)
        sigma2_hat = RSS / (out.fun.size - len(_init_params))
        cov = np.linalg.inv(J.T @ J)
        se = np.sqrt(np.diag(cov))

        params_val.append(out.x)
        params_err.append(se)
        covs.append(cov)
        res_vars.append(sigma2_hat)
        is_success.append(out.success)

    ax = core.AxisManager(tod.dets)
    pvals = np.array(params_val)  # shape (ndet, 3)
    perrs = np.array(params_err)  # shape (ndet, 3)
    ax.wrap('cx0',                 pvals[:,0],          [(0, 'dets')])
    ax.wrap('cy0',                 pvals[:,1],          [(0, 'dets')])
    ax.wrap('cr',                  pvals[:,2],          [(0, 'dets')])
    ax.wrap('cx0_err',             perrs[:,0],          [(0, 'dets')])
    ax.wrap('cy0_err',             perrs[:,1],          [(0, 'dets')])
    ax.wrap('cr_err',              perrs[:,2],          [(0, 'dets')])
    ax.wrap('covariance',          np.array(covs),      [(0, 'dets')])
    ax.wrap('residual_var',        np.array(res_vars),  [(0, 'dets')])
    ax.wrap('is_success',          np.array(is_success),[(0, 'dets')])
    tod.wg.wrap('cfit_result', ax)
    return fit_results
### Circle fitting functions to here ###

### Elliptical fitting functions start from here ###
def _ellipse_model(params, x, y, xerr, yerr):
    A, B, a, b, theta = params

    # Translate to center
    dx = x - A
    dy = y - B

    # Rotate
    cos_t = np.cos(theta)
    sin_t = np.sin(theta)
    x_rot =  dx * cos_t + dy * sin_t
    y_rot = -dx * sin_t + dy * cos_t

    # Implicit ellipse function F(x,y)=0
    m = (x_rot / a)**2 + (y_rot / b)**2 - 1.0

    # Gradients wrt observed x,y (needed for error propagation)
    inv_a2 = 1.0 / (a * a)
    inv_b2 = 1.0 / (b * b)
    Fx = 2.0 * (x_rot * inv_a2 * cos_t - y_rot * inv_b2 * sin_t)
    Fy = 2.0 * (x_rot * inv_a2 * sin_t + y_rot * inv_b2 * cos_t)

    # Error propagation: Var[m] ≈ Fx^2 * xerr^2 + Fy^2 * yerr^2
    sigma_m2 = Fx**2 * (xerr**2) + Fy**2 * (yerr**2)

    # Avoid division by zero (e.g., at exact boundary), add tiny floor
    eps = 1e-15
    chi = m / np.sqrt(sigma_m2 + eps)   # shape: (N,)

    return chi

def fit_with_ellipse(tod):

    _cal = tod.wg.cal_data
    fit_results = []
    params_val = []
    params_err = []
    covs       = []
    res_vars   = []
    is_success = []

    for i in range(tod.dets.count):
        obs = np.array([_cal.Q[:, i], _cal.U[:, i]])
        err = np.array([_cal.Qerr[:, i], _cal.Uerr[:, i]])

        # initial parameters
        A0 = np.nanmean(obs[0])
        B0 = np.nanmean(obs[1])
        a0 = max(np.sqrt(obs[0]**2 + obs[1]**2))
        b0 = min(np.sqrt(obs[0]**2 + obs[1]**2))
        theta0 = 0.0
        _init_params = np.array([A0, B0, a0, b0, theta0])
        _bounds = ([-2*a0, -2*b0, 0, 0, -np.pi/2], [2*a0, 2*b0, 10*a0, 10*b0, np.pi/2])

        # fit the data
        out = least_squares(
            _ellipse_model, _init_params, bounds=_bounds,
            args=(obs[0], obs[1], err[0], err[1])
        )
        fit_results.append(out)

        J = out.jac
        cov = np.linalg.inv(J.T @ J)
        params_val.append(out.x)
        params_err.append(np.sqrt(np.diag(cov)) * np.sqrt(out.fun.size - len(_init_params)))
        covs.append(cov)
        res_vars.append(np.sum(out.fun**2) / (len(out.fun) - len(_init_params)))
        is_success.append(out.success)

    # wrap results into the axis manager
    ax = core.AxisManager(tod.dets)
    vals = np.array(params_val)  # shape (ndet, 5)
    errs = np.array(params_err)  # shape (ndet, 5)
    ax.wrap('ex0',      vals[:,0], [(0, 'dets')])
    ax.wrap('ey0',      vals[:,1], [(0, 'dets')])
    ax.wrap('ea',       vals[:,2], [(0, 'dets')])
    ax.wrap('eb',       vals[:,3], [(0, 'dets')])
    ax.wrap('etheta',   vals[:,4], [(0, 'dets')])
    ax.wrap('ex0_err',  errs[:,0], [(0, 'dets')])
    ax.wrap('ey0_err',  errs[:,1], [(0, 'dets')])
    ax.wrap('ea_err',   errs[:,2], [(0, 'dets')])
    ax.wrap('eb_err',   errs[:,3], [(0, 'dets')])
    ax.wrap('etheta_err',errs[:,4], [(0, 'dets')])
    ax.wrap('covariance', np.array(covs),       [(0, 'dets')])
    ax.wrap('residual_var', np.array(res_vars), [(0, 'dets')])
    ax.wrap('is_success',   np.array(is_success),  [(0, 'dets')])
    tod.wg.wrap('efit_result', ax)
    return fit_results
### Elliptical fitting functions to here ###

# Get and wrap the calibration angle by the wire grid
def get_cal_gamma(tod, merge=True, remove_cal_data=False, num_bins=18, gap_size=np.deg2rad(5.)):
    """
    Calibrate detector polarization angle with a circle model

    Parameters
    ----------
        tod : AxisManager
        merge : bool (default, True)
        remove_cal_data : bool (defalut) False
        num_bins : int
            see _ignore_outlier_angle
        gap_size : float
            see _ignore_outlier_angle

    Returns
    -------
        ax : AxisManager
            which includes the calibrated angle of gamma in the sky coordinate, etc.

    Notes
    -----
        gamma_raw(_err) is the raw output of the calibration for each wire grid step.
        wires_relative_power is the signal intensity of the wires, which will relatively change depending on the ambient temperature.
        gamma(_err) is the main result of the calibration using Sparse Wire Grid.
        background_pol_rad or background_pol_relative_power is the Q/U-plane offsets of the detectors signal respect to the wires' reflection.
        theta_det_instr is gamma translated to the instrumental angle printed on the silicon wafer.

    """
    _cd = tod.wg.cal_data
    _cfr = tod.wg.cfit_result
    _det_angle = []
    _det_angle_err = []
    for _i in range(tod.wg.wg_steps.count):

        # remove the offset of the circle
        Qcal = _cd.Q.T[:,_i] - _cfr.cx0
        Ucal = _cd.U.T[:,_i] - _cfr.cy0
        _atan_sig = np.arctan2(Ucal, Qcal)%(2*np.pi)

        # 0.5*(2*theta_det + 2*theta_wire) - theta_wire
        _det_angle.append(
            (0.5*(_atan_sig[:]) - _cd.theta_wire_rad[_i,np.newaxis]%(2*np.pi))
        )
        _det_angle_err.append(
            np.sqrt(
                (_cd.Uerr.T[:,_i]**2 + _cd.Qerr.T[:,_i]**2) * 0.5
                + (_cfr.cy0_err[_i]**2 + _cfr.cx0_err[_i]**2) * 0.5 / _cfr.cr[_i]
            )
        )

    _det_angle = np.unwrap(np.array(_det_angle).T, period=np.pi)
    _det_angle_err = np.array(_det_angle_err).T

    # calibrated gamma
    gamma = np.nanmean(_det_angle, axis=1)%np.pi
    gamma_err = np.nanmean(_det_angle_err, axis=1)/np.sqrt(np.shape(_det_angle_err)[1])

    # back ground polarization
    _bg_theta = (0.5*np.arctan2(_cfr.cy0, _cfr.cx0)  - gamma)%np.pi
    _bg_amp = np.sqrt(_cfr.cx0**2 + _cfr.cy0**2)

    ax = core.AxisManager(tod.dets, tod.wg.wg_steps)
    ax.wrap('gamma_raw',                     _det_angle,        [(0, 'dets'), (1, 'wg_steps')])
    ax.wrap('gamma_raw_err',                 _det_angle_err,    [(0, 'dets'), (1, 'wg_steps')])
    ax.wrap('wires_relative_power',          _cfr.cr,           [(0, 'dets')])
    ax.wrap('gamma',                         gamma,             [(0, 'dets')])
    ax.wrap('gamma_err',                     gamma_err,         [(0, 'dets')])
    ax.wrap('background_pol_rad',            _bg_theta,         [(0, 'dets')])
    ax.wrap('background_pol_relative_power', _bg_amp,           [(0, 'dets')])
    ax.wrap('theta_det_instr',               0.5*np.pi - gamma, [(0, 'dets')]) # instumental angle of dets
    if remove_cal_data:
        tod.move('wg', None)
    if merge:
        tod.wrap('gamma_cal', ax)
    return ax

def get_ecal_gamma(tod):
    """
    Calibrate dtector polarization angle with an ellipse model.
    See also get_cal_gamma.
    """
    _cal = tod.wg.cal_data
    _efr = tod.wg.efit_result

    ang = []
    for _i in range(tod.wg.wg_steps.count):
        Qraw = (_cal.Q.T[:,_i] - _efr.ex0)
        Uraw = (_cal.U.T[:,_i] - _efr.ey0)

        Qrot = + Qraw * np.cos(_efr.etheta) + Uraw * np.sin(_efr.etheta)
        Urot = - Qraw * np.sin(_efr.etheta) + Uraw * np.cos(_efr.etheta)
        Qmod = Qrot /_efr.ea
        Umod = Urot /_efr.eb

        Qcal = Qmod * np.cos(_efr.etheta) - Umod * np.sin(_efr.etheta)
        Ucal = Qmod * np.sin(_efr.etheta) + Umod * np.cos(_efr.etheta)

        atan = np.arctan2(Ucal, Qcal)
        atan = atan%(2*np.pi)
        ang.append(
            0.5*atan[:] - _cal.theta_wire_rad[_i, np.newaxis]%(2*np.pi)
        )
    det_angle = np.unwrap(np.array(ang).T, period=np.pi)
    R = np.abs(np.nanmean(np.exp(1j*det_angle), axis=1))
    std_err = np.sqrt(-2*np.log(R))
    det_angle_err = np.repeat(np.atleast_2d(std_err), tod.wg.wg_steps.count, axis=0).T

    # gamma calibrated with ellipse model
    gamma = np.nanmean(det_angle, axis=1)%np.pi
    gamma_err = std_err

    ax = core.AxisManager(tod.dets, tod.wg.wg_steps)
    ax.wrap('gamma_raw',        det_angle%np.pi,   [(0, 'dets'), (1, 'wg_steps')])
    ax.wrap('gamma_raw_err',    det_angle_err,     [(0, 'dets'), (1, 'wg_steps')])
    ax.wrap('theta_det_instr',  0.5*np.pi - gamma, [(0, 'dets')])
    ax.wrap('gamma',            gamma,             [(0, 'dets')])
    ax.wrap('gamma_err',        gamma_err,        [(0, 'dets')])
    tod.wrap('gamma_ecal', ax)
    return ax



"""
Helper functions for Healpix operations

Functions are provided to transform between "full", "compressed", and "tiled"
Healpix maps. This follows a scheme in which the full map is broken up into
individual tiles defined by a lower resolution map. The setup ("geometry") is
defined by the base nside, nside_tile of the tiling, and the ordering. Tiled
maps have nTile = 12 * nside_tile**2, and must use "NEST" ordering in this
scheme. If no tiling is used, "RING" is allowed.

Individual tiles are also marked as "active" or "inactive" depending on
whether they include any non-zero pixels. This information is contained in an
active_tile_list: a len(nTile) boolean array of which tiles are active.

- Full map: A normal healpix map (in NEST ordering). Numpy arrays with shape
  (...,npix); can have any number of leading dimensions.
- Tiled map: A len(nTile) list of tiles. None for inactive tiles and a numpy
  array of shape (..., npix/nside) for active tiles.
- Compressed map: Numpy array of shape (nActiveTile, ..., npix/nside).
  Requires additional active tile information to recover a full map. The
  position of the tile axis in a compressed map can be controlled with the
  tile_axis_pos argument.
"""

import numpy as np
from types import SimpleNamespace

def get_geometry(nside, nside_tile=None, ordering='NEST'):
    """Make a simple object to contain Healpix geometry information.
    It has attributes nside, nside_tile, and ordering as given.
    If tiled, the additional attribute 'ntile' is automatically set to
    True (nside_tile='auto') or the number of tiles (nside_tile is an int).

    Args:
      nside: int >=1, power of 2. Nside of the base map.
      nside_tile: int or 'auto' for a tiled map. None for an un-tiled map.
      ordering: 'NEST' for tiled maps. 'NEST' or 'RING' for un-tiled.

    """
    hp_geom = SimpleNamespace(nside=nside, nside_tile=nside_tile, ordering=ordering)
    if nside_tile is not None:
        hp_geom.ntile = True
    if isinstance(nside_tile, (int, np.integer)):
        hp_geom.ntile = 12 * nside_tile**2

    check_geometry(hp_geom)
    return hp_geom

def get_active_tile_list(tiled):
    """From a tiled map, get a len(nTile) boolean array of which tiles are active"""
    return np.array([tile is not None for tile in tiled])

def full_to_compressed(full, active_tile_list, tile_axis_pos=0):
    """Get a compressed map from a full NEST map.

    Args:
      full: Full map in NEST.
      active_tile_list: len(nTile) boolean array of active tiles
      tile_axis_pos: Index for the tile axis in the compressed array
    """
    npix = full.shape[-1]
    npix_per_tile = int(npix / len(active_tile_list))
    cmap = []
    for tile_ind in range(len(active_tile_list)):
        if active_tile_list[tile_ind]:
            cmap.append(full[..., npix_per_tile * tile_ind : npix_per_tile * (tile_ind+1)])
    cmap = np.array(cmap, dtype=full.dtype)
    cmap = np.moveaxis(cmap, 0, tile_axis_pos)
    return cmap

def full_to_tiled(full, active_tile_list):
    """Get a tiled map from a full healpix map."""
    compressed = full_to_compressed(full, active_tile_list)
    tiled = compressed_to_tiled(compressed, active_tile_list)
    return tiled

def compressed_to_full(compressed, active_tile_list, tile_axis_pos=0):
    """Get a full NEST map from a compressed map. Empty tiles are filled with zeros.
    See full_to_compressed for args.
    """
    compressed = np.moveaxis(compressed, tile_axis_pos, 0)
    tile_shape = compressed[0].shape
    npix_per_tile = tile_shape[-1]
    super_shape = tile_shape[:-1]
    npix = len(active_tile_list) * npix_per_tile # ntiles * npix_per_tile
    out = np.zeros(super_shape + (npix,))
    tile_inds = [ii for ii in range(len(active_tile_list)) if active_tile_list[ii]]
    for ii in range(len(compressed)):
        tile_ind = tile_inds[ii]
        out[..., npix_per_tile * tile_ind : npix_per_tile * (tile_ind+1)] = compressed[ii]
    return out

def compressed_to_tiled(compressed, active_tile_list, tile_axis_pos=0):
    """Get a tiled map from a compressed map.
    See full_to_compressed for args.
    """
    compressed = np.moveaxis(compressed, tile_axis_pos, 0)
    out = [None] * len(active_tile_list)
    inds = np.where(active_tile_list)[0]
    assert compressed.shape[0] == inds.size
    for ii, ind in enumerate(inds):
        out[ind] = compressed[ii]
    return out

def tiled_to_compressed(tiled, tile_axis_pos=0):
    """Get a compressed map by removing Nones from a tiled map."""
    arr = np.array([tiled[ii] for ii in range(len(tiled)) if tiled[ii] is not None])
    return np.moveaxis(arr, 0, tile_axis_pos)

def tiled_to_full(tiled):
    """Get a full NEST map from a tiled map. Empty tiles are filled with zeros."""
    compressed = tiled_to_compressed(tiled)
    active_tile_list = get_active_tile_list(tiled)
    return compressed_to_full(compressed, active_tile_list)

def check_valid_nside(nside, name='nside'):
    if not isinstance(nside, (int, np.integer)):
        raise TypeError(f"{name} is of type {type(nside)}; should be an integer")
    if not ((nside >= 1) and ((nside & nside-1) == 0)): # Check for power of 2
        raise ValueError(f"Invalid {name} {nside}. Must be a power of 2 and >= 1")
    return True

def check_geometry(geom):
    # Check for required attributes
    for attr in ['nside', 'nside_tile', 'ordering']:
        if not hasattr(geom, attr):
            raise AttributeError(f"Healpix geometry missing required attribute {attr}")

    nside, nside_tile, ordering = geom.nside, geom.nside_tile, geom.ordering
    # Validate nside
    check_valid_nside(nside)

    # Validate nside_tile
    if (nside_tile is not None) and (nside_tile != "auto"):
        if isinstance(nside_tile, str):
            raise ValueError(f"nside_tile is invalid str {nside_tile}; should be integer, 'auto', or None")

        check_valid_nside(nside_tile, 'nside_tile')
        if nside_tile > nside:
            raise ValueError(f"Invalid nside_tile {nside_tile}; cannot be greated than nside {nside}")

    # Validate ordering
    if not (ordering in ['RING', 'NEST']):
        raise ValueError(f"Invalid value of 'ordering': {ordering}; should be 'RING' or 'NEST'")

    # Validate ntile
    is_tiled = (nside_tile is not None)
    ntile = hasattr(geom, 'ntile')
    if not ((is_tiled and ntile) or ((not is_tiled) and (not ntile))):
        raise ValueError("geom.ntile set incorrectly. Should be set for tiled and unset for un-tiled.")

    return True



################################
####### Noise model stuff ######
################################
import numpy as np
import so3g
from pixell import fft, utils, bunch

from .utils import *

class Nmat:
    def __init__(self):
        """Initialize the noise model. In subclasses this will typically set up parameters, but not
        build the details that depend on the actual time-ordered data"""
        self.ivar  = np.ones(1, dtype=np.float32)
        self.ready = True
    def build(self, tod, **kwargs):
        """Measure the noise properties of the given time-ordered data tod[ndet,nsamp], and
        return a noise model object tailored for that specific tod. The returned object
        needs to provide the .apply(tod) method, which multiplies the tod by the inverse noise
        covariance matrix. Usually the returned object will be of the same class as the one
        we call .build(tod) on, just with more of the internal state initialized."""
        return self
    def apply(self, tod):
        """Multiply the time-ordered data tod[ndet,nsamp] by the inverse noise covariance matrix.
        This is done in-pace, but the result is also returned."""
        self.check_ready()
        return tod*self.ivar
    def white(self, tod):
        """Like apply, but without detector or time correlations"""
        self.check_ready()
        return tod*self.ivar
    def write(self, fname):
        self.check_ready()
        bunch.write(fname, bunch.Bunch(type="Nmat"))
    @staticmethod
    def from_bunch(data): return Nmat()
    def check_ready(self):
        if not self.ready:
            raise ValueError("Attempt to use partially constructed %s. Typically one gets a fully constructed one from the return value of nmat.build(tod)" % type(self).__name__)

class NmatUncorr(Nmat):
    def __init__(self, spacing="exp", nbin=100, nmin=10, window=2, bins=None, ips_binned=None, ivar=None, nwin=None):
        self.spacing    = spacing
        self.nbin       = nbin
        self.nmin       = nmin
        self.bins       = bins
        self.ips_binned = ips_binned
        self.ivar       = ivar
        self.window     = window
        self.nwin       = nwin
        self.ready      = bins is not None and ips_binned is not None and ivar is not None

    def build(self, tod, srate, **kwargs):
        # Apply window while taking fft
        nwin  = utils.nint(self.window*srate)
        apply_window(tod, nwin)
        ft    = fft.rfft(tod)
        # Unapply window again
        apply_window(tod, nwin, -1)
        return self.build_fourier(ft, tod.shape[1], srate, nwin=nwin)

    def build_fourier(self, ftod, nsamp, srate, nwin=0):
        ps = np.abs(ftod)**2
        del ftod
        if   self.spacing == "exp": bins = utils.expbin(ps.shape[-1], nbin=self.nbin, nmin=self.nmin)
        elif self.spacing == "lin": bins = utils.expbin(ps.shape[-1], nbin=self.nbin, nmin=self.nmin)
        else: raise ValueError("Unrecognized spacing '%s'" % str(self.spacing))
        ps_binned  = utils.bin_data(bins, ps) / nsamp
        ips_binned = 1/ps_binned
        # Compute the representative inverse variance per sample
        ivar = np.zeros(len(ps))
        for bi, b in enumerate(bins):
            ivar += ips_binned[:,bi]*(b[1]-b[0])
        ivar /= bins[-1,1]-bins[0,0]
        return NmatUncorr(spacing=self.spacing, nbin=len(bins), nmin=self.nmin, bins=bins, ips_binned=ips_binned, ivar=ivar, window=self.window, nwin=nwin)

    def apply(self, tod, inplace=False, exp=1):
        self.check_ready()
        if inplace: tod = np.array(tod)
        if self.nwin > 0: apply_window(tod, self.nwin)
        ftod = fft.rfft(tod)
        self.apply_fourier(ftod, tod.shape[1], exp=exp)
        fft.irfft(ftod, tod)
        apply_window(tod, self.nwin)
        return tod

    def apply_fourier(self, ftod, nsamp, exp=1):
        self.check_ready()
        # Candidate for speedup in C
        for bi, b in enumerate(self.bins):
            ftod[:,b[0]:b[1]] *= (self.ips_binned[:,None,bi])**exp/nsamp
        # I divided by the normalization above instead of passing normalize=True
        # here to reduce the number of operations needed

    def white(self, tod, inplace=True):
        self.check_ready()
        if not inplace: tod = np.array(tod)
        apply_window(tod, self.nwin)
        tod *= self.ivar[:,None]
        apply_window(tod, self.nwin)
        return tod

    def write(self, fname):
        self.check_ready()
        data = bunch.Bunch(type="NmatUncorr")
        for field in ["spacing", "nbin", "nmin", "bins", "ips_binned", "ivar", "window", "nwin"]:
            data[field] = getattr(self, field)
        bunch.write(fname, data)

    @staticmethod
    def from_bunch(data):
        return NmatUncorr(spacing=data.spacing, nbin=data.nbin, nmin=data.nmin, bins=data.bins, ips_binned=data.ips_binned, ivar=data.ivar, window=data.window, nwin=data.nwin)

class NmatDetvecs(Nmat):
    def __init__(self, bin_edges=None, eig_lim=16, single_lim=0.55, mode_bins=[0.25,4.0,20],
            downweight=[], window=2, nwin=None, verbose=False, bins=None,
            D=None, V=None, iD=None, iV=None, s=None, ivar=None, bmin_eigvec=1000):
        # This is all taken from act, not tuned to so yet
        if bin_edges is None: bin_edges = np.array([
            0.16, 0.25, 0.35, 0.45, 0.55, 0.70, 0.85, 1.00,
            1.20, 1.40, 1.70, 2.00, 2.40, 2.80, 3.40, 3.80,
            4.60, 5.00, 5.50, 6.00, 6.50, 7.00, 8.00, 9.00, 10.0, 11.0,
            12.0, 13.0, 14.0, 16.0, 18.0, 20.0, 22.0,
            24.0, 26.0, 28.0, 30.0, 32.0, 36.5, 41.0,
            45.0, 50.0, 55.0, 65.0, 70.0, 80.0, 90.0,
            100., 110., 120., 130., 140., 150., 160., 170.,
            180., 190.
        ])
        self.bin_edges = bin_edges
        self.mode_bins = mode_bins
        self.eig_lim   = np.zeros(len(mode_bins))+eig_lim
        self.single_lim= np.zeros(len(mode_bins))+single_lim
        self.verbose   = verbose
        self.downweight= downweight
        self.bins = bins
        self.window = window
        self.nwin   = nwin
        self.bmin_eigvec = bmin_eigvec
        self.D, self.V, self.iD, self.iV, self.s, self.ivar = D, V, iD, iV, s, ivar
        self.ready      = all([a is not None for a in [D, V, iD, iV, s, ivar]])

    def build(self, tod, srate, **kwargs):
        # Apply window before measuring noise model
        nwin  = utils.nint(self.window*srate)
        apply_window(tod, nwin)
        ftod  = fft.rfft(tod)
        # Unapply window again
        apply_window(tod, nwin, -1)
        return self.build_fourier(ftod, tod.shape[1], srate, nwin=nwin, **kwargs)

    def build_fourier(self, ftod, nsamp, srate, nwin=0, **kwargs):
        ndet, nfreq = ftod.shape
        dtype       = utils.real_dtype(ftod.dtype)
        # First build our set of eigenvectors in two bins. The first goes from
        # 0.25 to 4 Hz the second from 4Hz and up
        mode_bins = makebins(self.mode_bins, srate, nfreq, nmin=self.bmin_eigvec, rfun=np.ceil, cap=False)
        if np.any(np.diff(mode_bins) < 0):
            raise RuntimeError(f"At least one of the frequency bins has a negative range: \n{mode_bins}")
        # Then use these to get our set of basis vectors
        vecs = find_modes_jon(ftod, mode_bins, eig_lim=self.eig_lim, single_lim=self.single_lim, verbose=self.verbose)
        nmode= vecs.shape[1]
        if vecs.size == 0: raise errors.ModelError("Could not find any noise modes")
        # Cut bins that extend beyond our max frequency
        bin_edges = self.bin_edges[self.bin_edges < srate/2 * 0.99]
        bins      = makebins(bin_edges, srate, nfreq, nmin=5, rfun=np.round)
        nbin      = len(bins)
        # Now measure the power of each basis vector in each bin. The residual
        # noise will be modeled as uncorrelated
        E  = np.zeros([nbin,nmode])
        D  = np.zeros([nbin,ndet])
        Nd = np.zeros([nbin,ndet])
        for bi, b in enumerate(bins):
            # Skip the DC mode, since it's it's unmeasurable and filtered away
            b = np.maximum(1,b)
            E[bi], D[bi], Nd[bi] = measure_detvecs(ftod[:,b[0]:b[1]], vecs)
        # Optionally downweight the lowest frequency bins
        if self.downweight != None and len(self.downweight) > 0:
            D[:len(self.downweight)] /= np.array(self.downweight)[:,None]
        # Instead of VEV' we can have just VV' if we bake sqrt(E) into V
        V = vecs[None]*E[:,None]**0.5
        # At this point we have a model for the total noise covariance as
        # N = D + VV'. But since we're doing inverse covariance weighting
        # we need a similar representation for the inverse iN. The function
        # woodbury_invert computes iD, iV, s such that iN = iD + s iV iV'
        # where s usually is -1, but will become +1 if one inverts again
        iD, iV, s = woodbury_invert(D, V)
        # Also compute a representative white noise level
        bsize = bins[:,1]-bins[:,0]
        ivar  = np.sum(iD*bsize[:,None],0)/np.sum(bsize)
        # What about units? I haven't applied any fourier unit factors so far,
        # so we're in plain power units. From the uncorrelated model I found
        # that factor of tod.shape[1] is needed
        iD   *= nsamp
        iV   *= nsamp**0.5
        ivar *= nsamp
        # Fix dtype
        bins = np.ascontiguousarray(bins.astype(np.int32))
        D    = np.ascontiguousarray(D.astype(dtype))
        V    = np.ascontiguousarray(V.astype(dtype))
        iD   = np.ascontiguousarray(iD.astype(dtype))
        iV   = np.ascontiguousarray(iV.astype(dtype))

        return NmatDetvecs(bin_edges=self.bin_edges, eig_lim=self.eig_lim, single_lim=self.single_lim,
                window=self.window, nwin=nwin, downweight=self.downweight, verbose=self.verbose,
                bins=bins, D=D, V=V, iD=iD, iV=iV, s=s, ivar=ivar)

    def apply(self, tod, inplace=True, slow=False):
        self.check_ready()
        if not inplace: tod = np.array(tod)
        apply_window(tod, self.nwin)
        ftod = fft.rfft(tod)
        self.apply_fourier(ftod, tod.shape[1], slow=slow)
        fft.irfft(ftod, tod)
        apply_window(tod, self.nwin)
        return tod

    def apply_fourier(self, ftod, nsamp, slow=False):
        self.check_ready()
        dtype= utils.real_dtype(ftod.dtype)
        if slow:
            for bi, b in enumerate(self.bins):
                # Want to multiply by iD + siViV'
                ft    = ftod[:,b[0]:b[1]]
                iD    = self.iD[bi]/nsamp
                iV    = self.iV[bi]/nsamp**0.5
                ft[:] = iD[:,None]*ft + self.s*iV.dot(iV.T.dot(ft))
        else:
            so3g.nmat_detvecs_apply(ftod.view(dtype), self.bins, self.iD, self.iV, float(self.s), float(nsamp))
        # I divided by the normalization above instead of passing normalize=True
        # here to reduce the number of operations needed

    def white(self, tod, inplace=True):
        self.check_ready()
        if not inplace: tod = np.array(tod)
        apply_window(tod, self.nwin)
        tod *= self.ivar[:,None]
        apply_window(tod, self.nwin)
        return tod

    def write(self, fname):
        self.check_ready()
        data = bunch.Bunch(type="NmatDetvecs")
        for field in ["bin_edges", "eig_lim", "single_lim", "window", "nwin", "downweight",
                "bins", "D", "V", "iD", "iV", "s", "ivar"]:
            data[field] = getattr(self, field)
        try:
            bunch.write(fname, data)
        except Exception as e:
            msg = f"Failed to write {fname}: {e}"
            raise RuntimeError(msg)

    @staticmethod
    def from_bunch(data):
        return NmatDetvecs(bin_edges=data.bin_edges, eig_lim=data.eig_lim, single_lim=data.single_lim,
                window=data.window, nwin=data.nwin, downweight=data.downweight,
                bins=data.bins, D=data.D, V=data.V, iD=data.iD, iV=data.iV, s=data.s, ivar=data.ivar)

# Combination of NmatDetvecs and NmatUncorr. The first handles the correlations while the
# latter handles the variance
class NmatScaledvecs(Nmat):
    def __init__(self, nmat_uncorr, nmat_detvecs, window=0, nwin=0):
        self.nmat_uncorr  = nmat_uncorr
        self.nmat_detvecs = nmat_detvecs
        self.window       = window
        self.nwin         = nwin
    @property
    def ready(self): return self.nmat_uncorr.ready and self.nmat_detvecs.ready
    @property
    def ivar(self): return self.nmat_uncorr.ivar
    def build(self, tod, srate, **kwargs):
        # Apply window before measuring noise model
        nsamp = tod.shape[1]
        nwin  = utils.nint(self.window*srate)
        apply_window(tod, nwin)
        ftod  = fft.rfft(tod)
        nmat_uncorr = self.nmat_uncorr.build_fourier(ftod, nsamp, srate)
        # Whiten the tod before building the correlated model
        nmat_uncorr.apply_fourier(ftod, nsamp, exp=0.5)
        nmat_detvecs= self.nmat_detvecs.build_fourier(ftod, nsamp, srate)
        # Unapply window again
        apply_window(tod, nwin, -1)
        return NmatScaledvecs(nmat_uncorr, nmat_detvecs, window=self.window, nwin=self.nwin)

    def apply(self, tod, inplace=True):
        self.check_ready()
        if not inplace: tod = np.array(tod)
        apply_window(tod, self.nwin)
        ftod = fft.rfft(tod)
        # Our model is N" = Nu"**0.5 Nd" Nu**0.5. But since both models are
        # fourier-diagonal, this is equal to Nd" Nu"
        self.nmat_uncorr .apply_fourier(ftod, tod.shape[1])
        self.nmat_detvecs.apply_fourier(ftod, tod.shape[1])
        fft.irfft(ftod, tod)
        apply_window(tod, self.nwin)
        return tod

    def white(self, tod, inplace=True):
        self.check_ready()
        return self.nmat_uncorr.white(tod, inplace=inplace)

    def write(self, fname):
        self.check_ready()
        raise NotImplementedError

    @staticmethod
    def from_bunch(data):
        raise NotImplementedError

class NmatWhite(Nmat):
    def __init__(self, ivar=None):
        """
        This is a white noise model for the mapmaker.
        The white noise model is characterized by 
        (1) no correlations between detectors 
        (2) a white (flat) spectrum per detector  
        (3) the weights, if not passed in through ivar, are computed simply
        from the inverse variance of the timestream.

        Parameters
        ----------
        ivar : numpy.ndarray or None, optional
            Overwrite the inverse variance per detector

        Returns
        -------
        noise_model : An Nmat object with the noise model

        """
        self.ivar = ivar
    def build(self, tod, srate, **kwargs):
        if np.any(np.logical_not(np.isfinite(tod))):
            raise ValueError(f"There is a nan when calculating the white noise !!!")
        ivar = 1/np.var(tod,1)
        return NmatWhite(ivar)
    def apply(self, tod, inplace=False):
        self.check_ready()
        if inplace: tod = np.array(tod)
        tod *= self.ivar[:,None]
        return tod
    def white(self, tod, inplace=True):
        self.check_ready()
        return self.apply(tod, inplace=inplace)
    def write(self, fname):
        self.check_ready()
        data = bunch.Bunch(type="NmatWhite")
        for field in ["ivar"]:
            data[field] = getattr(self, field)
        bunch.write(fname, data)
    @staticmethod
    def from_bunch(data):
        return NmatWhite(ivar=data.ivar)
    @property
    def ready(self):
        return self.ivar is not None

class NmatUnit(Nmat):
    """
    This is a noise model that does nothing, equivalent to multiply by a 
    unit noise matrix
    """

    def __init__(self, ivar=None):
        self.ivar  = ivar
        self.ready = ivar is not None
    def build(self, tod, **kwargs):
        ndet, nsamps = tod.shape
        ivar = np.ones(ndet)
        return NmatUnit(ivar=ivar)
    def apply(self, tod):
        # the tod is returned intact
        self.check_ready()
        return tod
    def white(self, tod):
        # the tod is returned intact
        self.check_ready()
        return tod
    def write(self, fname):
        self.check_ready()
        data = bunch.Bunch(type="NmatUnit")
        for field in ["ivar"]:
            data[field] = getattr(self, field)
        bunch.write(fname, data)
    @staticmethod
    def from_bunch(data): 
        return NmatUnit(ivar=data.ivar)

def write_nmat(fname, nmat):
    nmat.write(fname)

def read_nmat(fname):
    data = bunch.read(fname)
    typ  = data.type.decode()
    if   typ == "NmatDetvecs": return NmatDetvecs.from_bunch(data)
    elif typ == "NmatUncorr":  return NmatUncorr .from_bunch(data)
    elif typ == "NmatWhite":   return NmatWhite  .from_bunch(data)
    elif typ == "NmatUnit":    return NmatUnit   .from_bunch(data)
    elif typ == "Nmat":        return Nmat       .from_bunch(data)
    else: raise IOError("Unrecognized noise matrix type '%s' in '%s'" % (str(typ), fname))



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Timestream flagging operations.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_simple_jumpcorrect(operators):
    """Add commandline args and operators for jump flagging.
    Args:
        operators (list):  The list of operators to extend.
    Returns:
        None
    """
    operators.append(toast.ops.SimpleJumpCorrect(name="simple_jumpcorrect", enabled=False))


@workflow_timer
def simple_jumpcorrect(job, otherargs, runargs, data):
    """Apply simple jump correction
    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.
    Returns:
        None
    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.simple_jumpcorrect.enabled:
        job_ops.simple_jumpcorrect.apply(data)


def setup_simple_deglitch(operators):
    """Add commandline args and operators for glitch flagging.
    Args:
        operators (list):  The list of operators to extend.
    Returns:
        None
    """
    operators.append(toast.ops.SimpleDeglitch(name="simple_deglitch", enabled=False))


@workflow_timer
def simple_deglitch(job, otherargs, runargs, data):
    """Apply simple deglitching
    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.
    Returns:
        None
    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.simple_deglitch.enabled:
        job_ops.simple_deglitch.apply(data)


def setup_flag_sso(operators):
    """Add commandline args and operators for SSO flagging.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.FlagSSO(name="flag_sso", enabled=False))


@workflow_timer
def flag_sso(job, otherargs, runargs, data):
    """Flag solarsystem objects.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.flag_sso.enabled:
        job_ops.flag_sso.detector_pointing = job_ops.det_pointing_azel
        job_ops.flag_sso.apply(data)


def setup_flag_noise_outliers(operators):
    """Add commandline args and operators for flagging detectors with bad noise.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.NoiseEstim(
            name="noise_cut",
            out_model="noise_cut",
            lagmax=1024,
            nbin_psd=64,
            nsum=1,
            naverage=64,
            enabled=False,
        )
    )
    operators.append(
        toast.ops.FitNoiseModel(
            name="noise_cut_fit",
            out_model="noise_cut_fit",
            enabled=True,
        )
    )
    operators.append(
        toast.ops.FlagNoiseFit(
            name="noise_cut_flag",
            sigma_NET=5.0,
            sigma_fknee=5.0,
            enabled=True,
        )
    )


@workflow_timer
def flag_noise_outliers(job, otherargs, runargs, data):
    """Flag detectors with extremely bad noise properties.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()
    timer = toast.timing.Timer()
    timer.start()

    # Configured operators for this job
    job_ops = job.operators

    # If any operators are disabled, just return
    all_enabled = (
        job_ops.noise_cut.enabled
        and job_ops.noise_cut_fit.enabled
        and job_ops.noise_cut_flag.enabled
    )
    if not all_enabled:
        log.info_rank(
            "Noise-based cut of detectors disabled", comm=data.comm.comm_world
        )
        return

    # Estimate noise.
    log.info_rank(
        "  Running noise estimation on raw data...", comm=data.comm.comm_world
    )
    job_ops.noise_cut.apply(data)
    log.info_rank(
        "  Estimated raw data noise in", comm=data.comm.comm_world, timer=timer
    )

    # Create a fit to this noise model
    job_ops.noise_cut_fit.noise_model = job_ops.noise_cut.out_model
    log.info_rank(
        "  Running 1/f fit to estimated raw noise model", comm=data.comm.comm_world
    )
    job_ops.noise_cut_fit.apply(data)
    log.info_rank(
        "  Fit raw 1/f noise model in", comm=data.comm.comm_world, timer=timer
    )

    # Flag detector outliers
    job_ops.noise_cut_flag.noise_model = job_ops.noise_cut_fit.out_model
    log.info_rank(
        "  Running flagging of noise model outliers...", comm=data.comm.comm_world
    )
    job_ops.noise_cut_flag.apply(data)
    log.info_rank(
        "  Flag raw noise model outliers in", comm=data.comm.comm_world, timer=timer
    )

    # Delete these temporary models
    toast.ops.Delete(
        meta=[
            job_ops.noise_cut.out_model,
            job_ops.noise_cut_fit.out_model,
        ]
    ).apply(data)


def setup_flag_diff_noise_outliers(operators):
    """Add commandline args and operators for flagging white noise outliers.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.SignalDiffNoiseModel(
            name="diff_noise_cut",
            noise_model="diff_noise_cut",
            enabled=False,
        )
    )
    operators.append(
        toast.ops.FlagNoiseFit(
            name="diff_noise_cut_flag",
            sigma_NET=5.0,
            enabled=True,
        )
    )


@workflow_timer
def flag_diff_noise_outliers(job, otherargs, runargs, data):
    """Flag detectors with outlier white noise properties.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()
    timer = toast.timing.Timer()
    timer.start()

    # Configured operators for this job
    job_ops = job.operators

    # If any operators are disabled, just return
    all_enabled = job_ops.diff_noise_cut.enabled and job_ops.diff_noise_cut_flag.enabled
    if not all_enabled:
        log.info_rank(
            "Noise-based cut of detectors disabled", comm=data.comm.comm_world
        )
        return

    # Estimate noise.
    log.info_rank(
        "  Running noise estimation on raw data...", comm=data.comm.comm_world
    )
    job_ops.diff_noise_cut.apply(data)
    log.info_rank(
        "  Estimated raw data noise in", comm=data.comm.comm_world, timer=timer
    )

    # Flag detector outliers
    job_ops.diff_noise_cut_flag.noise_model = job_ops.diff_noise_cut.noise_model
    log.info_rank(
        "  Running flagging of noise model outliers...", comm=data.comm.comm_world
    )
    job_ops.diff_noise_cut_flag.apply(data)
    log.info_rank(
        "  Flag raw noise model outliers in", comm=data.comm.comm_world, timer=timer
    )

    # Delete these temporary models
    toast.ops.Delete(
        meta=[
            job_ops.diff_noise_cut.noise_model,
        ]
    ).apply(data)


def setup_processing_mask(operators):
    """Add commandline args and operators for processing mask flagging.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.PixelsHealpix(
            name="processing_mask_pixels",
            pixels="pixels_processing_mask",
            enabled=False,
        )
    )
    operators.append(toast.ops.ScanHealpixMask(name="processing_mask", enabled=False))


@workflow_timer
def processing_mask(job, otherargs, runargs, data):
    """Raise data processing flags based on a mask.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.processing_mask.enabled:
        if job_ops.processing_mask_pixels.enabled:
            # We are using a custom pointing matrix
            job_ops.processing_mask_pixels.detector_pointing = (
                job_ops.det_pointing_radec
            )
            job_ops.processing_mask.pixel_dist = "processing_mask_pixel_dist"
            job_ops.processing_mask.pixel_pointing = job_ops.processing_mask_pixels
        else:
            # We are using the same pointing matrix as the mapmaking
            job_ops.processing_mask.pixel_dist = job_ops.binner.pixel_dist
            job_ops.processing_mask.pixel_pointing = job.pixels_solve
        job_ops.processing_mask.save_pointing = otherargs.full_pointing
        job_ops.processing_mask.apply(data)
